#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <vector>
#include "hash.h"
#include "lp_lib.h"
#include "structs.h"
#include "glocal.h"


#define MAX2(a,b) ((a)>(b))?(a):(b)
#define MIN2(a,b) ((a)<(b))?(a):(b)

#define MATE(x) (((x)%2)?((x)+1):((x)-1))
#define BASE(x) ((x+1)>>1)

#define MAX_INT ((1<<31)-1)

#define M1 1
#define M2 2
#define DM 3

#define SEQ_NAME_LEN 256
#define COLS_EDGE 6

#define SCOREFILE "/home/genome/lib/global.score"

//external array

extern vector<Fragment>fragments;

// Some utilities

void* ck_malloc(int size) {
  void *data = malloc(size);
  if (!data) {
    fprintf(stderr, "Out of memory\n");
    assert(0);
  }
  return data;
}

void* ck_realloc(void* ptr, int size) {
  void *data = realloc(ptr, size);
  if (!data) {
    fprintf(stderr, "Out of memory\n");
    assert(0);
  }
  return data;
}

// The structures

typedef struct alignment_end {
  int myrow;     // my row in LP matrix
  int links[2];  // the 2 adjacent alignment ends (Do I need them?)
  int actives[2]; // the 2 edges for flow solving
} align_end;

typedef struct alignment {
  int id;   
  int seqs[2];
  int starts[2]; 
  int ends[2];
  float chain_score[2]; // scores for chains w/ outlier genomes
  float score;  //score of the alignment
  char strand;
  char type;   //M1, M2, or DM
} align;


typedef struct alignment_joiner {
  int prev; //align_end before the join
  int next; //align_end after the join
  int mycol; //column of this edge in the LP matrix
  int padding[2]; //amount to pad after prev(0) and before next(1) for chain[4] and chain[5]
  float chain[COLS_EDGE]; //scores for middle(00), next-middle(10),  prev-middle(01) ,all(11), prev (4), next (5)
  float probs[COLS_EDGE]; //probs for middle(00), next-middle(10), prev-middle(01) and all(11)
  char usage;  // how the edge is used: not (00), next-edge (10), edge-prev(01) or through (11)
  char genome; // in which genome the link is
  char clean;
} edge;

typedef struct align_graph {
  int numaligns;
  int numedges;

  align* aligns; // 1...numaligns BASE(end) gets the align
  align_end* ends; // 1...2*numaligns. MATE(end) gets opp end
  edge* edges; //1...numedges edges used in flow solving
} graph;


typedef struct string_mapper { //mapper from chrom names to numbers
  int     numseqs;
  int     listsize;
  hash_t* myhash;
  char**   mylist;
  int*    sorted; // if sorted[i] > sorted[j] then i > j
} str_map;

typedef struct outlier_hit { // hit to an outlier genome
  char seqs[2][257];
  int starts[2];
  int ends[2];
  float score;
  char strand; // "+" or "-"
  char genome; // on which of my two genomes this hit is
} ohit;

typedef struct myfile {
  int cntr;
  int size;
  ohit* list;
  float myscore;
} filer;

typedef struct outgroup {
  char* name;
  float weight;
} outlier;

//the statics 

graph mygraph; //The graph

outlier* outliers; //outlier genomes
int numouts; 
float* out_weights;
int* sorted_aligns;

char* genome[2];

char* typestr[4] = {"", "M1", "M2", "DM"};

str_map seqmap[2];

char prevgenome; //genome in which the previous ohit was
int Which_genome; //which genome I am currently working on, to avoid passing it everywhere
                  // as a parameter.
int curr_out;  //the current outlier we are dealing with

//output utils 

void print_ohit(ohit* what) {
  return;
  fprintf(stderr, "OH %s %d %d; %s %d %d; score = %f (%c)\n", what->seqs[0], what->starts[0], what->ends[0],
	  what->seqs[0], what->starts[1], what->ends[1], what->score, what->strand);   
}

void print_align(align* what) {
  //  return;
  fprintf(stderr, "AL %s %d %d; %s %d %d; score = %f (%c) %s\n", seqmap[0].mylist[what->seqs[0]], 
	  what->starts[0], what->ends[0],  seqmap[1].mylist[what->seqs[1]], what->starts[1], 
	  what->ends[1], what->score, what->strand, typestr[what->type]);   
}

void print_edge(edge* what) {
  return;
  fprintf(stderr, "EDGE(%d,%d)\n", what->usage, what->genome);
  print_align(&(mygraph.aligns[BASE(what->prev)]));
  print_align(&(mygraph.aligns[BASE(what->next)]));
}

// comparator functions

static int strmap_comp(const void* p1, const void* p2) {
  int i = *((int*)p1);
  int j = *((int*)p2);
  return strcmp(seqmap[Which_genome].mylist[i], seqmap[Which_genome].mylist[j]);
}

static int graph_comp(const void* p1, const void* p2) {
  int i = *((int*)p1);
  int j = *((int*)p2);

  if (seqmap[Which_genome].sorted[mygraph.aligns[i].seqs[Which_genome]] > 
      seqmap[Which_genome].sorted[mygraph.aligns[j].seqs[Which_genome]])
    return 1;
  if (seqmap[Which_genome].sorted[mygraph.aligns[i].seqs[Which_genome]] < 
      seqmap[Which_genome].sorted[mygraph.aligns[j].seqs[Which_genome]])
    return -1;

  if (mygraph.aligns[i].starts[Which_genome] > mygraph.aligns[j].ends[Which_genome])
    return 1;
  if (mygraph.aligns[j].starts[Which_genome] > mygraph.aligns[i].ends[Which_genome])
    return -1;
  //  fprintf (stderr, "WARNING: OVERLAPPING INTERVALS: [%d,%d], [%d,%d]\n", 
  //	   mygraph.aligns[i].starts[Which_genome], mygraph.aligns[i].ends[Which_genome],
  //	   mygraph.aligns[j].starts[Which_genome], mygraph.aligns[j].ends[Which_genome]);
  if (mygraph.aligns[i].starts[Which_genome] > mygraph.aligns[j].starts[Which_genome])
    return 1;
  if (mygraph.aligns[i].starts[Which_genome] < mygraph.aligns[j].starts[Which_genome])
    return -1;

  return 0;
}

// Score function
float scorefunc(float a1, float a2, float u, float i, int prev_active, int curr_active, float* savr) {
  float mx = MAX2(a1, a2);
  float mn = MIN2(a1, a2);
  float sc = 0;

  //  fprintf(stderr, "SC call %f, %f, %f, %f\n", a1,a2,u,i);
  fflush(stderr);

  if (prev_active) {
    sc += mygraph.aligns[prev_active].score;
  }
  if (curr_active) {
    sc += mygraph.aligns[curr_active].score;
  }
  
  sc = sc / 2.0;

  if ((mn - i) > 0.0000001) {
    *savr =  ((u-mx)/(mn-i));
    //
    //    fprintf(stderr, "u %f, mx %f, mn %f, i %f, 1 %f 2 %f = %f\n",u,mx,mn,i,a1,a2, *savr); 

    if (*savr < 0) {
      //      fprintf(stderr, "WW1: u %f, mx %f, mn %f, i %f, 1 %f 2 %f = %f\n",u,mx,mn,i,a1,a2, *savr); 
      *savr = 0;
    }
    if (*savr > 1) {
      //      fprintf(stderr, "WW2: u %f, mx %f, mn %f, i %f, 1 %f 2 %f = %f\n",u,mx,mn,i,a1,a2, *savr); 
      *savr = 1;
    }
    (*savr) *= outliers[curr_out].weight;
    sc *= (*savr);

    //    fprintf(stderr, "SC = %f, CO = %f svr = %f\n", sc, outliers[curr_out].weight, *savr);
    return sc;
  }
  *savr = 0;
  return 0;
}

// Actual functions

void init_mapper() {
  int i;
  for (i=0; i < 2; i++) {
    seqmap[i].numseqs = 0;
    seqmap[i].listsize = 1;
    seqmap[i].myhash = (hash_t*)ck_malloc(sizeof(hash_t));
    hash_init(seqmap[i].myhash, 32);
    seqmap[i].mylist = (char**)ck_malloc(sizeof(char*) * seqmap[i].listsize);
    seqmap[i].sorted = (int*)ck_malloc(sizeof(int) * seqmap[i].listsize);
  }
}

void close_mapper() {
  int i,j;
  int* temp;
  for (i=0; i < 2; i++) {
    temp = (int*) ck_malloc(sizeof(int) * seqmap[i].numseqs);
    seqmap[i].mylist = (char**)ck_realloc(seqmap[i].mylist, sizeof(char*) * seqmap[i].numseqs);
    seqmap[i].sorted = (int*)ck_realloc(seqmap[i].sorted, sizeof(int) * seqmap[i].numseqs);
    for (j=0; j < seqmap[i].numseqs; j++) {
      temp[j] = j;
    }
    Which_genome = i;     //sort temp based on strings
    qsort (temp, seqmap[i].numseqs,sizeof(int), strmap_comp);
    for (j=0; j < seqmap[i].numseqs; j++) {
      seqmap[i].sorted[temp[j]] = j;
    }
    free(temp);
  }
}

int read_align(FILE* afile, align* tgt) {
  char buffer[1025];
  char sname[2][SEQ_NAME_LEN];
  char typestr[256];
  int  res;
  char *name; 
  int i;
    
  if (!fgets(buffer, 1024, afile)) {
    //    fprintf(stderr, "Done reading\n");
    return 0;
  }
  //  fprintf(stderr, "Still reading\n");
  sscanf(buffer, "%d %s %d %d %s %d %d %c %f %s", &(tgt->id), sname[0], &(tgt->starts[0]), &(tgt->ends[0]),
	 sname[1], &(tgt->starts[1]), &(tgt->ends[1]), &(tgt->strand), &(tgt->score), typestr); 

  tgt->chain_score[0] =  tgt->chain_score[1] = 0;

  if (!strcmp(typestr, "M1")) {   tgt->type = 1;  } 
  else if (!strcmp(typestr, "M2")) {   tgt->type = 2;  } 
  else if (!strcmp(typestr, "DM")) {   tgt->type = 3;  } 
  else { fprintf(stderr, "Bad align type: %s\n", typestr); assert(0);}

  for (i=0; i < 2; i++) {
    res = hash_lookup(seqmap[i].myhash, sname[i]);
    if (res >= 0) {
      tgt->seqs[i] = res;
    }
    else {
      name = (char*) ck_malloc (strlen(sname[i])+1);
      strcpy (name, sname[i]);
      seqmap[i].mylist[seqmap[i].numseqs] = name;
      hash_insert(seqmap[i].myhash, name, seqmap[i].numseqs);
      tgt->seqs[i] = seqmap[i].numseqs++;
      if (seqmap[i].numseqs >= seqmap[i].listsize) {
	seqmap[i].mylist = (char**)ck_realloc(seqmap[i].mylist, sizeof(char*) * seqmap[i].listsize * 2);
	seqmap[i].sorted = (int*)ck_realloc(seqmap[i].sorted, sizeof(int) * seqmap[i].listsize * 2);
	seqmap[i].listsize *= 2;
      }
    }
  }
  return 1;
}

void init_ends(int which) {
  mygraph.ends[which*2].links[0] =  mygraph.ends[which*2].links[1] = 0;
  mygraph.ends[which*2].actives[0] =  mygraph.ends[which*2].actives[1] = 0;
  mygraph.ends[which*2-1].links[0] =  mygraph.ends[which*2-1].links[1] = 0;
  mygraph.ends[which*2-1].actives[0] =  mygraph.ends[which*2-1].actives[1] = 0;
}

int initialize_graph(FILE* alignfile) {
  
  int align_size = 2;
  int flag;

  init_mapper();

  mygraph.numaligns = 0;
  mygraph.numedges = 0;
  mygraph.aligns = (align*) ck_malloc(sizeof(align) * align_size);
  mygraph.ends = (align_end*) ck_malloc(sizeof(align_end) * (align_size*2));
  mygraph.edges = (edge*) ck_malloc(sizeof(edge)*2);

  while (!feof(alignfile)) {
    flag = read_align(alignfile, &(mygraph.aligns[mygraph.numaligns+1]));
    if (!flag) {
      //      fprintf(stderr,"Breaking\n");
      break;
    }
    mygraph.numaligns++;
    init_ends(mygraph.numaligns);
    if (mygraph.numaligns >= (align_size - 1)) {
      mygraph.aligns = (align*) ck_realloc(mygraph.aligns, sizeof(align)*(2*align_size));
      mygraph.ends = (align_end*) ck_realloc(mygraph.ends, sizeof(align_end)*(4*align_size));
      align_size *= 2;
    }
  }
  close_mapper();
  //  fprintf(stderr, "Done init, %d\n", mygraph.numaligns);
}

int inside(int i, int o) {
  if (seqmap[Which_genome].sorted[mygraph.aligns[i].seqs[Which_genome]] != 
      seqmap[Which_genome].sorted[mygraph.aligns[o].seqs[Which_genome]])
    return 0;
  if (mygraph.aligns[o].starts[Which_genome] > mygraph.aligns[i].starts[Which_genome])
    return 0;
  if (mygraph.aligns[o].ends[Which_genome] < mygraph.aligns[i].ends[Which_genome])
    return 0;
  return 1;

}

int sort_graph (int which) {
  int i, size = mygraph.numaligns;
  int nelem = 0, currelem=0;
  for (i = 1; i <= size; i++) {
    if (mygraph.aligns[i].type & which) {
      sorted_aligns[nelem++] = i;
    }
    //    fprintf(stderr, "type=%d which=%d ne=%d\n", mygraph.aligns[i].type, which, nelem);
  }
  Which_genome = which-1;
  qsort(sorted_aligns, nelem, sizeof(int), graph_comp);
  for (i = 1; i < nelem; i++) {
    if (inside(sorted_aligns[i], sorted_aligns[currelem])) {
      fprintf(stderr, "%d inside %d\n", sorted_aligns[i], sorted_aligns[currelem]); 
      sorted_aligns[i] = -1; 
    }
    else
      currelem = i;
  }
  currelem = 0;

  for (i=0; i < nelem; i++) {
    if (sorted_aligns[i] == -1) {
      currelem = MAX2(i, currelem); 
      while (currelem < nelem && sorted_aligns[currelem] == -1) {
	currelem++;
      }
      if (currelem == nelem) {
	break; //nothing to do, i is the size of resulting sorted_aligns
      }
      else {
	sorted_aligns[i] = sorted_aligns[currelem];
	sorted_aligns[currelem] = -1;
      }
    }
  }
  //  assert(0);
  return i;
}

int callnum = 0;

int read_hit(FILE* inp, ohit* tgt) {
  int i, w = -1, retval = -1;
  char temp[1025];
  char genm[2][256];

  if (!fgets(temp, 1024, inp))
    return 0;
  
  if (temp[0] == '#') {
    fprintf(stderr, "cn = %d\n", callnum);
    sscanf(temp, "# %s %s", genm[0], genm[1]);

    for (i = 0; i < numouts; i++) {
      if (!strcmp(genm[1], outliers[i].name)) {
	w = i;
	break;
      }
    }
    if (i== numouts) {
      fprintf(stderr, "Error reading input 1\n");
      exit(2);
    }
    curr_out = w;

    for (i=0; i < 2; i++) {
      if (!strcmp(genm[0], genome[i])) {
	w = i;
	break;
      }
    }
    fprintf(stderr, "%s %s %d %d\n",genm[0], genome[w], w, prevgenome);
    if (i == 2) {
      fprintf(stderr, "Error reading input 2\n");
      exit(2);
    }




    if (w != prevgenome) { //switched genomes or first call
      prevgenome = w;
      if (w == callnum) {
	//	callnum++; //callnum was equal so continue reading
	return -1;
      }
      else if (w > callnum) {
	callnum++;  //they weren't so go to next sequence
	return 0;
      }
      else {
	fprintf(stderr, "error in input, expected base %d, but got a smaller one %d\n", callnum, w);
	exit(2);
      }
    }


    if (callnum > 3) {
      fprintf(stderr, "Error reading input 3\n");
      exit(2);
    }
    return -1; //return the current base genome
  }

 end:
  sscanf (temp, "%s %d %d %s %d %d %c %f\n", &(tgt->seqs[0]), &(tgt->starts[0]), &(tgt->ends[0]),
	  &(tgt->seqs[1]), &(tgt->starts[1]), &(tgt->ends[1]), &(tgt->strand), &(tgt->score));
  tgt->genome = prevgenome;
  //  fprintf(stderr, "read %s",temp);
  return 1;
}


filer prev_file;
filer curr_file;
filer edge_file;

int sizeedges = 2;

int init_filers() {
  prev_file.cntr = curr_file.cntr = edge_file.cntr = 0;
  prev_file.size = curr_file.size = edge_file.size = 0;
  prev_file.list = curr_file.list = edge_file.list = NULL;
  prev_file.myscore = curr_file.myscore = edge_file.myscore = 0.0;
}

int fresh_filer(filer* target) { //refresh the target filer
  target->cntr = 0;
  target->size = 0;
  free (target->list);
  target->list = NULL;
}

int add_hit(filer* where, ohit* what) {
  if (where->cntr >= where->size) {
    where->size = where->size * 2 + 5;
    where->list = (ohit*) realloc(where->list, sizeof(ohit) * where->size);
  }
  where->list[where->cntr] = *what;
  (where->cntr)++;
}

Fragment createFrag(ohit temp)
{
  Fragment frag;
  frag.seq1Start= temp.starts[0];
  frag.seq1End= temp.ends[0];

  frag.seq2Start= temp.starts[1];;
  frag.seq2End= temp.ends[1];
  frag.printFlag = FALSE;

  strcpy(frag.seq1Name,temp.seqs[0]); 
  strcpy(frag.seq2Name,temp.seqs[1]); 

  if(temp.strand=='+')
    frag.strand = POSITIVE;
  else if (temp.strand =='-')
    frag.strand = NEGATIVE;
  else
    {
      printf("Incorrect value for strand type/n");
      exit(0);
    }

  frag.score= temp.score;
  frag.back= NULL;
  frag.totalScore = 0;
  return frag;
}

float runGlocal(filer** a, int numelem) {

  int i, j, tlen=0;
  
  //  fprintf(stderr,"runglocal\n");

  for (i=0; i < numelem; i++) {
    tlen += a[i]->cntr;
  }
  fragments.clear();
  fragments.reserve(tlen);
  for (i=0; i < numelem; i++) {
    for (j=0; j < a[i]->cntr; j++) {
      fragments.push_back(createFrag(a[i]->list[j]));
    }
  }
  //  fprintf(stderr, "anout %d\n", i);
  float g =  do_chain();
  //  fprintf(stderr, "done\n");
  return g;
}

float get_score1(filer* a) {
  filer* t[1];
  t[0] = a;
  return runGlocal(t,1);
}

float get_score2(filer* a, filer* b) {
  filer* t[2];
  t[0] = a;
  t[1] = b;
  return runGlocal(t,2);
}

float get_score3(filer* a, filer* b, filer *c) {
  filer* t[3];
  t[0] = a;
  t[1] = b;
  t[2] = c;
  return runGlocal(t,3);
}


int set_scores(int firstedge, int lastedge, int prev_active, int curr_active, int which) {
  int edgenum;
  float tcount = lastedge-firstedge+1;
  float allscores[6]; //1, 2, 3, 12, 23, 123 = 1 prev, 2 edge, 3 curr
  float savr;
  int i;


  //  fprintf(stderr, "counts: %d %d %d\n", prev_file.cntr, edge_file.cntr, curr_file.cntr);
  //compute & set scores 
  //  fprintf(stderr, "Setting scores %d %d\n", prev_active, curr_active);


  if (prev_active) {
    allscores[0] = prev_file.myscore;
    allscores[3] = get_score2(&prev_file, &edge_file);
    //    fprintf(stderr, "3.1=%f\n", allscores[3]);
  }
  allscores[1] = get_score1(&edge_file);
  if (curr_active) {
    allscores[2] = get_score1(&curr_file);
    curr_file.myscore = allscores[2];
    allscores[4] = get_score2(&curr_file, &edge_file);
  }
  if (prev_active && curr_active)
    allscores[5] = get_score3(&curr_file, &prev_file, &edge_file);
  
  for (edgenum = firstedge; edgenum <= lastedge; edgenum++) {
    //    fprintf(stderr, "3.2=%f\n", allscores[3]);
    //    fprintf(stderr, "4.2=%f\n", allscores[4]);
    //    fprintf(stderr, "5.2=%f\n", allscores[5]);

    mygraph.edges[edgenum].clean = 1;
    //    fprintf(stderr, "here\n");
    if (prev_active && edgenum == firstedge) {
      //      fprintf(stderr, "3.3=%f\n", allscores[3]);
      mygraph.edges[edgenum].chain[1] += scorefunc(allscores[0], allscores[1], allscores[3], 0, 
						  prev_active, curr_active, &savr);
      mygraph.edges[edgenum].probs[1] += savr;
    }
    else if (edgenum == firstedge) {
      /*      mygraph.edges[edgenum].chain[1] = -999999;
	      mygraph.edges[edgenum].probs[1] = -999999;
	      mygraph.edges[edgenum].chain[0] = -999999;
	      mygraph.edges[edgenum].probs[0] = -999999;
      */
    }

    if (curr_active && edgenum == lastedge) {
      //      fprintf(stderr, "4.3=%f\n", allscores[4]);
      mygraph.edges[edgenum].chain[2] += scorefunc(allscores[2], allscores[1], allscores[4], 0, 
						  prev_active, curr_active, &savr);
      mygraph.edges[edgenum].probs[2] += savr;
    }
    else if (edgenum == lastedge) {
      /*      mygraph.edges[edgenum].chain[2] = -999999;
	      mygraph.edges[edgenum].probs[2] = -999999;
	      mygraph.edges[edgenum].chain[0] = -999999;
	      mygraph.edges[edgenum].probs[0] = -999999;
      */
    }
    
    if (prev_active && curr_active) {
      //      fprintf(stderr, "5.3=%f\n", allscores[5]);
      mygraph.edges[edgenum].chain[0] += scorefunc(allscores[3], allscores[4], allscores[5], allscores[1], 
						  BASE(mygraph.edges[edgenum].prev), 
						  BASE(mygraph.edges[edgenum].next), &savr)/tcount;
      mygraph.edges[edgenum].probs[0] += savr/tcount;
    }

    mygraph.edges[edgenum].chain[4] = 0; // TODO for gap division
    mygraph.edges[edgenum].probs[4] = 0;
    mygraph.edges[edgenum].chain[5] = 0;
    mygraph.edges[edgenum].probs[5] = 0;
  }
  
  /*
  fprintf(stderr, "[");
  for (i = 0; i < 6; i++)
    fprintf(stderr, " %f", allscores[i]);
  fprintf(stderr, "]=[");
  for (i = 0; i < 3; i++)
    fprintf(stderr, " %f(%f)", mygraph.edges[firstedge].chain[i], mygraph.edges[firstedge].probs[i]);
  fprintf(stderr, "]\n");
  */
}


int score_edge(int prev_active, int curr_active, int which) {
  filer tfile;
  int firstedge=0, lastedge=0;

  if (prev_active) {
    if (which && (mygraph.aligns[prev_active].strand == '-')) {
      firstedge = mygraph.ends[2*prev_active-1].actives[which];
    }
    else {
      firstedge = mygraph.ends[2*prev_active].actives[which];
    }
  }
  if (curr_active) {
    if (which && (mygraph.aligns[curr_active].strand == '-')) {
      lastedge = mygraph.ends[2*curr_active].actives[which];
    }
    else {
      lastedge = mygraph.ends[2*curr_active-1].actives[which];
    }
  }
  if (!firstedge)
    firstedge = lastedge;
  if (!lastedge)
    lastedge = firstedge;

  //  fprintf(stderr,"score\n");
  if (firstedge || lastedge)
    set_scores(firstedge, lastedge, prev_active, curr_active, which);
  //  fprintf(stderr,"donescore\n");
  tfile = prev_file;
  prev_file = curr_file;
  curr_file = tfile;
  fresh_filer(&curr_file);
  fresh_filer(&edge_file);
  //  fprintf(stderr,"refresh!\n");
}

// -1 if hit before align, 0 if overlapping, 1 if after
int direction(ohit* myhit, int al, int which) {
  
  int t;

  t= strcmp (myhit->seqs[0], seqmap[which].mylist[mygraph.aligns[al].seqs[which]]);
  
  if (t)
    return t;

  if (myhit->starts[0] >= mygraph.aligns[al].starts[which] &&
      myhit->ends[0] <= mygraph.aligns[al].ends[which]) //inside
    return 0;
  if ((myhit->starts[0] <= mygraph.aligns[al].starts[which] && 
      myhit->ends[0] >= mygraph.aligns[al].starts[which]) ||

      (myhit->starts[0] <= mygraph.aligns[al].ends[which] && 
       myhit->ends[0] >= mygraph.aligns[al].ends[which])) //overlapping
    return 0;

  if (myhit->ends[0] < mygraph.aligns[al].starts[which])
    return -1;
  if (myhit->starts[0] > mygraph.aligns[al].ends[which])
    return 1;
  
  //  fprintf(stderr, "I shouldn't be here (direction)\n");
  assert(0);
}

int get_weights(int which, int ssize) {
  int i = 0, rhres, tt;
  int prev_active = 0;
  int curr_active = 0; 
  int edgenum = mygraph.numedges;
  char prevseq[SEQ_NAME_LEN] = "";

  ohit currhit = { {{'\0'}, {'\0'}}, {0, 0}, {0, 0}, 0.0, '\0', '\0'};
  //  print_align(&(mygraph.aligns[1]));
  fprintf(stderr, "INto get weights %d\n", mygraph.numedges);

  while (!feof(stdin)) {

    rhres = read_hit(stdin, &currhit);
    // if no more data or changed seqs close anything open
    //    fprintf(stderr, "strngs: %s %s\n", prevseq, currhit.seqs[0]);

    if (rhres <= 0 || strcmp(prevseq, currhit.seqs[0])) {
      fprintf(stderr, "change %d %s %s\n", rhres, prevseq, currhit.seqs[0]);
      if (curr_active) {
	score_edge(prev_active, curr_active, which);
	prev_active = curr_active;
	curr_active = 0;
	score_edge(prev_active, curr_active, which);
	//	fprintf(stderr," Closed EDGE1\n");
      }
      else if (prev_active) {
	score_edge(prev_active, curr_active, which);
	//	fprintf(stderr," Closed EDGE2\n");
      }
      //      fprintf(stderr, "here %s %s\n", prevseq, currhit.seqs[0]);
      else {
	fresh_filer(&edge_file);
      }

      prev_active = curr_active = 0;
      if (rhres <= 0) { //done with a pair
	return rhres; //negative means we loop back to this func, zero quits.
      }
      strcpy(prevseq, currhit.seqs[0]);
    }

    //  print_align(&(mygraph.aligns[1]));
  begin: //if we finished with an align we adjust variables and re-process the hit

    if (curr_active) {
      tt = direction(&currhit, curr_active, which);
      if (!tt) {
	//	fprintf(stderr, "HIT\n");
	//	print_ohit(&currhit);
	//	print_align(&(mygraph.aligns[curr_active]));
	add_hit(&curr_file, &currhit);
      }
      else if (tt < 0) {
	//	fprintf(stderr, "REVERSAL\n");
	//	print_ohit(&currhit);
	//	print_align(&(mygraph.aligns[curr_active]));
	add_hit(&edge_file, &currhit);
      }
      else {
	score_edge(prev_active, curr_active, which);
	//	fprintf(stderr," Closed EDGE3\n");
	prev_active = curr_active;
	curr_active = 0;
	i++;
	if (! (i % 100)) {
	  fprintf(stderr, "\nALGN=%d\n", i);
	}
	goto begin;
      }
    }
    else {
      if (i >= ssize || direction(&currhit, sorted_aligns[i], which) < 0) {
	//	fprintf(stderr, "IN GAP\n");
	add_hit(&edge_file, &currhit);
      }
      else if (direction(&currhit, sorted_aligns[i], which) == 0) {
	curr_active = sorted_aligns[i];
	//	fprintf(stderr, "curractive=%d, i=%d\n", curr_active, i);
	//	print_align(&(mygraph.aligns[sorted_aligns[i]]));
	goto begin;
      }
      else {
	while (i < ssize && direction(&currhit, sorted_aligns[i], which) > 0) {
	  //	  fprintf (stderr, "skipping [%d]=%d...\n", i, sorted_aligns[i]);
	  //	  print_ohit(&currhit);
	  //	  print_align(&(mygraph.aligns[sorted_aligns[i]]));
	  i++; //skip aligns if we had no hits on them]
	  if (! (i % 100)) {
	    fprintf(stderr, "\nALGN=%d\n", i);
	  }
	}
	goto begin;
      }
    }
  }
  fprintf(stderr, "Numedges = %d\n", edgenum);
  return 0;
}

int close_edge(int edgenum, int prev_active, int curr_active, int which) {
  int i;
  //  setup

  //  fprintf(stderr, "Called close %d %d %d %d\n", edgenum, prev_active,curr_active, which);
  //  fprintf(stderr, "ADDING EDGE\n");

  //  print_align(&mygraph.aligns[prev_active]);
  //  print_align(&mygraph.aligns[curr_active]);

  if (prev_active) {
    if (which && (mygraph.aligns[prev_active].strand == '-')) {
      mygraph.ends[2*prev_active-1].actives[which] = edgenum;
      mygraph.edges[edgenum].prev = 2*prev_active - 1;
    }
    else {
      mygraph.ends[2*prev_active].actives[which] = edgenum;
      mygraph.edges[edgenum].prev = 2*prev_active;
    }
  }
  else {
    mygraph.edges[edgenum].prev = 0;
  }

  if (curr_active) {
    if (which && (mygraph.aligns[curr_active].strand == '-')) {
      mygraph.ends[2*curr_active].actives[which] = edgenum;
      mygraph.edges[edgenum].next = 2*curr_active;
    }
    else {
      mygraph.ends[2*curr_active-1].actives[which] = edgenum;
      mygraph.edges[edgenum].next = 2*curr_active-1;
    }
  }
  else {
    mygraph.edges[edgenum].next = 0;
  }

  mygraph.edges[edgenum].mycol = -1;
  mygraph.edges[edgenum].usage = 0;
  mygraph.edges[edgenum].genome = which;
  mygraph.edges[edgenum].clean = 0;
  mygraph.edges[edgenum].padding[0] = mygraph.edges[edgenum].padding[1] = 0;
  for (i=0; i < COLS_EDGE; i++) {
    mygraph.edges[edgenum].chain[i] = 0;
    mygraph.edges[edgenum].probs[i] = 0;
  }

  
  //  grow
  prev_active = curr_active;
  if (edgenum >= sizeedges-1) {
    mygraph.edges = (edge*) ck_realloc(mygraph.edges, sizeof(edge) * (sizeedges *= 2));
  }
}


void setup_edges(int which, int ssize) {
  int i=0;
  int edgenum = mygraph.numedges;
  int prev = 0;
  int tnext, next;

  if (ssize == 0)
    return;

  next = sorted_aligns[i];

  
  for (i=1; i < ssize; i++) {
    edgenum++;
    close_edge(edgenum, prev, next, which);
    tnext = sorted_aligns[i];
    if (mygraph.aligns[next].seqs[which] !=  mygraph.aligns[tnext].seqs[which]) {
      edgenum++;
      close_edge(edgenum, next, 0, which);
      prev = 0;
      next = tnext;
    }
    else {
      prev = next;
      next = tnext;
    }
  }
  edgenum++;
  close_edge(edgenum, sorted_aligns[ssize-1], 0, which);
  mygraph.numedges = edgenum;
}


int add_columns(lprec * mylp, int currcol, int mye) {
  int i;
  REAL vals[3];
  int cols[3];
  
  //  fprintf(stderr, "Adding column %d\n", currcol);

  for (i=0; i < COLS_EDGE; i++) {
    if (!add_columnex(mylp,0,0,0)) {
      fprintf (stderr, "Failed to add column\n");
      assert(0);
    }
    set_int(mylp, (currcol-1)*COLS_EDGE+1 + i, 1);
    set_bounds(mylp, (currcol-1)*COLS_EDGE+1 + i, 0, 1);
    set_obj(mylp, (currcol-1)*COLS_EDGE+1 + i, mygraph.edges[mye].chain[i]);
  }



  vals[0] = vals[1] = vals[2] = 1;
  cols[0] = (currcol-1)*COLS_EDGE + 2;
  cols[1] = (currcol-1)*COLS_EDGE + 3;
  cols[2] = (currcol-1)*COLS_EDGE + 6;
  add_constraintex (mylp, 3, vals, cols, EQ, 1);

}

int do_component(int e) {
  // edge is column, node is row

  int currcol = 1; 
  int columns[11];
  REAL values[11];
  int i = e, j;
  int o, mx;
  int prevend = mygraph.edges[i].next;
  int currend = mygraph.edges[i].prev;
  int k, n, t;
  int* col2edge = (int*) ck_malloc(sizeof(int)*2);
  REAL* row;
  int c2esize = 2;
  int flag;

  lprec * mylp = make_lp(0,0);

  set_outputstream (mylp, stderr);
  set_verbose(mylp, SEVERE);
  //walk left all the way, marking the nodes and adding them.

  add_columns(mylp, currcol,i);
  mygraph.edges[i].mycol = currcol++;
  col2edge[currcol-1] = i;
  if (currcol >= c2esize) {
    col2edge = (int*) ck_realloc(col2edge, sizeof(int)* (c2esize *= 2));
  }

  
  for (n = 0; n < 2; n++) {
    //    fprintf(stderr, "Walking %d\n", n);
    while (currend) {
      mx = 2;
      flag = 0;
      if (mygraph.ends[currend].actives[0] == i) {
	o = mygraph.ends[currend].actives[1];
      }
      else if (mygraph.ends[currend].actives[1] == i) {
	o = mygraph.ends[currend].actives[0];
      }
      else {
	fprintf (stderr, "Bad structures\n");
	assert (0);
      }
      //      fprintf(stderr, "curredge from->to\n");
      //      print_align(&(mygraph.aligns[BASE(currend)]));
      //      print_align(&(mygraph.aligns[BASE(prevend)]));
      //      fprintf(stderr, "oedge is %d\n", o);      
      if (!o) { //no other edge
	mx = 1;
      }
      if (!o || mygraph.edges[o].mycol >= 0) {
	//	fprintf(stderr, "marked next edge: %d\n",mygraph.edges[o].mycol);
	flag = 1;
      }
      else {
	add_columns(mylp, currcol, o);
	mygraph.edges[o].mycol = currcol++;
	col2edge[currcol-1] = o;
	if (currcol >= c2esize) {
	  col2edge = (int*) ck_realloc(col2edge, sizeof(int)* (c2esize *= 2));
	}	
      }
      //add balalnce 
      t = i;
      for (k = 0; k < mx; k++) {
	//	fprintf(stderr, "adding constraint %d\n",mygraph.edges[t].mycol);

	columns[3*k+0] = COLS_EDGE * (mygraph.edges[t].mycol - 1) + 1;
	if (currend == mygraph.edges[t].prev) {
	  columns[3*k+1] = COLS_EDGE * (mygraph.edges[t].mycol - 1) + 2;
	  columns[3*k+2] = COLS_EDGE * (mygraph.edges[t].mycol - 1) + 4;
	}
	else {
	  columns[3*k+1] = COLS_EDGE * (mygraph.edges[t].mycol - 1) + 3;
	  columns[3*k+2] = COLS_EDGE * (mygraph.edges[t].mycol - 1) + 5;
	}
	values[3*k+0] = 1;
	values[3*k+1] = 1;
	values[3*k+2] = 1;
	t = o;
      }
      add_constraintex(mylp, 3*mx, values, columns, EQ, 1);
      
      if (flag) {
	//	fprintf(stderr, "Breaking in DC\n");
	break;
      }
      i = o;
      prevend = currend;
      if (mygraph.edges[i].prev == prevend)
	currend = mygraph.edges[i].next;
      else if (mygraph.edges[i].next == prevend)
	currend = mygraph.edges[i].prev;
      else {
	fprintf(stderr, "bad structures2\n");
	assert (0);
      }
    }
    prevend = mygraph.edges[e].prev;
    currend = mygraph.edges[e].next;
    i = e;
  }

  // solve LP
  //  print_lp(mylp);
  
  set_maxim(mylp);
  if (solve(mylp)) {
    fprintf(stderr, "Couldn't solve the LP\n");
    print_lp(mylp);
    assert(0);
  }
  else {
    //    print_solution(mylp, currcol*COLS_EDGE);
    //    fprintf(stderr, "\n");
  }

  // get results for edges
  row = (REAL*) ck_malloc (sizeof(REAL) * get_Ncolumns(mylp));
  get_variables(mylp, row);

  for (i = 1; i < currcol; i++) {

    for (j=0; j < 3; j++) {
      if (row[(i-1)*COLS_EDGE + j] !=0 && row[(i-1)*COLS_EDGE + j] != 1)
	fprintf (stdout, "# non integral!!! %d, %d, %f\n",i,j,row[(i-1)*COLS_EDGE + j]);
    }

    //
    if (!mygraph.edges[col2edge[i]].prev || !mygraph.edges[col2edge[i]].next)
      mygraph.edges[col2edge[i]].usage = 0;
    //
    else if (row[(i-1)*COLS_EDGE + 0] == 1)
      mygraph.edges[col2edge[i]].usage = 3;
    else if (row[(i-1)*COLS_EDGE + 1] == 1)
      mygraph.edges[col2edge[i]].usage = 1; 
    else if (row[(i-1)*COLS_EDGE + 2] == 1)
      mygraph.edges[col2edge[i]].usage = 2;
    else
      mygraph.edges[col2edge[i]].usage = 0;
    //    fprintf(stderr, "col2edge %d\n", col2edge[i]);
    print_edge(&(mygraph.edges[col2edge[i]]));
  }


  delete_lp(mylp);
  free (col2edge);
  free (row);
}


int solve_flow() {
  int i;
  
  for (i = 1; i <= mygraph.numedges; i++) {
    // take an edge.
    if (mygraph.edges[i].mycol >= 0) {
      //      fprintf(stderr, "%d marked %d", i, mygraph.edges[i].mycol);
      continue;
    }
    //    fprintf(stderr, "About to do component %d\n", i);
    do_component(i);
  }

  for (i = 1; i <= mygraph.numedges; i++) {
    mygraph.edges[i].mycol = -1;
  }
}

int output_align(int e, int whflag) {
  int drawend = (mygraph.edges[e].usage == 1)? mygraph.edges[e].prev: mygraph.edges[e].next;

  char strnd = '+';
  if (whflag == 1)
    strnd = '-';

  fprintf(stdout, "%d %s %c\n", mygraph.aligns[BASE(drawend)].id, genome[mygraph.edges[e].genome], strnd);
  
}

void output_edge(int e, int whflag) {
  //usage == whflag means before align, ow after align

  int prevend;
  int nextend;
  int goodend;
  int gen = mygraph.edges[e].genome;
  int lastend;
  int mystart;
  char strnd = '+';
  int t = 1;

  if (whflag == 2) {
    prevend = mygraph.edges[e].prev;
    nextend = mygraph.edges[e].next;
    lastend = 0;
    if (prevend) {
      mystart = MAX2(mygraph.aligns[BASE(prevend)].ends[gen], mygraph.aligns[BASE(prevend)].starts[gen]);
    }
    else {
      mystart = 0;
    }
    if (nextend)
      lastend = MIN2(mygraph.aligns[BASE(nextend)].ends[gen], mygraph.aligns[BASE(nextend)].starts[gen]);
    //    fprintf(stdout, "here1 [%d %d]%c\n", mystart, lastend, strnd);
    else
      lastend = MAX_INT;
  }

  else if (whflag == 1) {
    strnd = '-';
    nextend = mygraph.edges[e].prev;
    prevend = mygraph.edges[e].next;
    if (prevend) {
      lastend = MIN2(mygraph.aligns[BASE(prevend)].ends[gen], mygraph.aligns[BASE(prevend)].starts[gen]);
    }
    else {
      lastend = MAX_INT; //END OF chromosome
    }
    if (nextend)
      mystart = MAX2(mygraph.aligns[BASE(nextend)].ends[gen], mygraph.aligns[BASE(nextend)].starts[gen]);
    else {
      mystart = 0;
    }
    //    fprintf(stdout, "here2 [%d %d]%c\n", mystart, lastend, strnd);
  }
  else {
    assert(0);
  }

  goodend = (nextend)? nextend:prevend;

  if (mystart < (lastend-t)) {
    if (lastend == MAX_INT)
      lastend = 0; //Alex wants a -1 to mark the end of a chromosome
    fprintf(stdout, "%s %s %d %d %c   # (%f %f %f)\n", genome[gen], 
	    seqmap[gen].mylist[mygraph.aligns[BASE(goodend)].seqs[gen]], 
	    mystart+t, lastend-t, strnd,
	    mygraph.edges[e].probs[0],mygraph.edges[e].probs[1], mygraph.edges[e].probs[2]);
  }
  else if (lastend < (mystart-1))
    return;
}

int output_link(int e, int whflag) {
  int prevend = (whflag == 2)? mygraph.edges[e].prev : mygraph.edges[e].next;
  int nextend = (whflag == 1)? mygraph.edges[e].prev : mygraph.edges[e].next;
  char strnd = '+';
  if (whflag == 1)
    strnd = '-';
  
  fprintf(stdout, "%d %d %s %c # (%f %f %f) [%d %d]\n", mygraph.aligns[BASE(prevend)].id, 
	  mygraph.aligns[BASE(nextend)].id, genome[mygraph.edges[e].genome], strnd,
	  mygraph.edges[e].probs[0],mygraph.edges[e].probs[1],
	  mygraph.edges[e].probs[2], prevend, nextend);
}
  
int output_tigs() {
  int i, j, ne, pe, w, t;
  int whflag, prevwhflag;
  int curredge, prevedge, opend;
  int currend, prevend;

  int minedge = -1;
  int prevmin, minend, minwhflag;

  for (i = 1; i <= mygraph.numedges; i++) {
    // take an edge.
    if ((mygraph.edges[i].mycol >= 0) || !mygraph.edges[i].usage) {
      continue;
    }
    // Follow backwards to one end
    minedge = -1;
    curredge = i;
    currend =  mygraph.edges[curredge].prev;
    prevend =  mygraph.edges[curredge].next;
    whflag = 1;
    opend = 0;

    while (currend && (mygraph.edges[curredge].usage & whflag)) {  
      //      fprintf(stderr, "Roll back\n");
      print_edge(&mygraph.edges[curredge]);
      if (mygraph.edges[curredge].mycol == -2) {
	//cycle
	mygraph.edges[minedge].mycol = 0; //break the cycle
	fprintf(stdout, "# broke cycle; whflag = %d, minwhflag = %d\n", whflag, minwhflag);
	if (minedge == curredge) {
	  prevmin = prevedge;
	  minwhflag = prevwhflag;
	}
	curredge = prevmin;
	whflag = minwhflag;
	if (mygraph.edges[prevmin].prev == MATE(minend)) {
	  prevend = mygraph.edges[prevmin].next;
	  currend = mygraph.edges[prevmin].prev;
	}
	else if (mygraph.edges[prevmin].next == MATE(minend)) {
	  prevend = mygraph.edges[prevmin].prev;
	  currend = mygraph.edges[prevmin].next;
	}
	else {
	  fprintf(stderr, "Unrecoverable error (screwup cycle): %d (%d %d) (%d %d)\n", 
		  minend, mygraph.edges[prevmin].prev, mygraph.edges[prevmin].next,
		  mygraph.edges[minedge].prev, mygraph.edges[minedge].next);
	  assert(0);
	}
	break;
	// is bug here whflag is not set ???
      }

      if (minedge < 0 || mygraph.edges[curredge].probs[0] < mygraph.edges[minedge].probs[0]) {
	minedge = curredge; 
	prevmin = prevedge;
	minend = prevend;
	minwhflag = prevwhflag;
      }
      
      mygraph.edges[curredge].mycol = -2;
      //      assert(mygraph.aligns[BASE(currend)].id != 385);
      opend = MATE(currend);
      ne = -1;

      prevwhflag = whflag;
      for (j=0; j < 2; j++) {
	w = mygraph.ends[opend].actives[j];
	if ((mygraph.edges[w].prev == opend) && (mygraph.edges[w].usage & 1)) {
	  assert(ne < 0);
	  ne = w;
	  whflag = 2;
	}
	if ((mygraph.edges[w].next == opend) && (mygraph.edges[w].usage & 2)) {
	  assert(ne < 0);
	  ne = w;
	  whflag = 1;
	}
      }
      if (ne > 0) {
	prevedge = curredge;
	curredge = ne; 
	prevend = opend;
	currend = 
	  (opend==mygraph.edges[curredge].prev)? mygraph.edges[curredge].next : mygraph.edges[curredge].prev; 
      }
      else {
	fprintf(stdout, "# another break; ne = %d whflag = %d, minwhflag = %d\n", ne, whflag, minwhflag);      
	//	assert(mygraph.aligns[BASE(opend)].id != 385);

	break;
      }
    }

    if (minedge < 0) {
      fprintf(stdout, "# skipped all; whflag = %d, minwhflag = %d\n", whflag, minwhflag);      
    }

    //swap prevend & currend, since we now walk in opposite direction
    opend = prevend;    prevend = currend;    currend = opend;
    //reverse whflag
    whflag = (whflag == 1)? 2 : 1;

    // print & mark all edges along the path to other end

    w = 0;
    pe = 0;
    while (1) { 
      //      fprintf(stderr, "Roll forward\n");
      // print_edge(&mygraph.edges[curredge]);
      if (mygraph.edges[curredge].mycol == 0) {
	break;
      }
      if (mygraph.edges[curredge].usage == 1 || mygraph.edges[curredge].usage == 2) {
	if ((mygraph.edges[curredge].usage != whflag) && !pe) {
	  output_align(curredge, whflag);
	  pe = curredge;
	}	  
	output_edge(curredge, whflag);
      }
      else {
	pe = curredge; 
	output_link(curredge, whflag);
      }
      mygraph.edges[curredge].mycol = 0; //marking
      
      if (!currend || !(mygraph.edges[curredge].usage & whflag)) {  
	break;
      }

      opend = MATE(currend);
      ne = 0;
      for (j=0; j < 2; j++) {
	t = mygraph.ends[opend].actives[j];
	if ((mygraph.edges[t].prev == opend) && (mygraph.edges[t].usage & 1)) {
	  assert (!ne);
	  ne = t;
	  whflag = 2;
	}
	if ((mygraph.edges[t].next == opend) && (mygraph.edges[t].usage & 2)) {
	  assert (!ne);
	  ne = t;
	  whflag = 1;
	}
      }
      if (ne) {
	curredge = ne; 
	prevend = opend;
	currend = 
	  (opend==mygraph.edges[curredge].prev)? mygraph.edges[curredge].next : mygraph.edges[curredge].prev; 
      }
      else {
	break;
      }
    }
    if (!pe) {
      output_align(curredge, whflag);
    }
    fprintf(stdout, "#\n");
  }
}


int read_input(int argc, char** argv, int first_param, char* score_file) {
  
  int i;
  numouts = (argc - (4 + first_param - 1))/2;
  prevgenome = -1;
  genome[0] = argv[1 + first_param - 1];
  genome[1] = argv[2 + first_param - 1];

  outliers = (outlier*) ck_malloc(sizeof(outlier) * numouts);

  for (i = 0; i < numouts; i++) {
    outliers[i].name = argv[2*i + 4 + first_param - 1];
    outliers[i].weight = atof(argv[2*i + 5 + first_param - 1]);
    fprintf(stderr, "outlier %d, %s = %f\n", i, outliers[i].name, outliers[i].weight);
  }

  
  
  init_glocal(score_file);
  //TODO when we have weights we should get those properly
}

void dump_weights() {
  int i, j, t;

  fprintf(stderr, "into dump\n");

  for (i=1; i < mygraph.numedges; i++) {
    //    fprintf(stderr, "i = %d, t = %d\n", i,t);
    if (mygraph.edges[i].clean) {
      fprintf(stdout, "%d", i);
      for (j=0; j < COLS_EDGE; j++) {
	fprintf(stdout, " %f %f",mygraph.edges[i].chain[j], mygraph.edges[i].probs[j]);
      }
      fprintf(stdout, "\n");
    }  
  }
}

void read_weights() {
  int j, e, q;
  float r, s;
  fprintf(stderr, "reading\n");

  while (!feof(stdin)) {
    if (fscanf(stdin, "%d", &e)<1) {
      if (feof(stdin))
	break;
      fprintf (stderr, "parse error 1\n");
      exit(1);
    }
    for (j=0; j < COLS_EDGE; j++) {
      if ((q = fscanf(stdin, " %f %f", &r, &s)) < 2) {
	fprintf (stderr, "parse error 2[%d, %d, %d]\n",e,q,j);
	exit(1);
      }
      //      fprintf(stderr, "set %d %f %f\n",e,r,s);
      mygraph.edges[e].chain[j] += r;
      mygraph.edges[e].probs[j] += s;
      mygraph.edges[e].clean += 1;
    }
  }

  fprintf(stderr, "check\n");
  for (j = 1; j <= mygraph.numedges; j++) {
    if (! (mygraph.edges[j].clean)) {
      e = 0;
      if (!mygraph.edges[j].prev) {
	mygraph.edges[j].chain[1] = -999999;
	mygraph.edges[j].probs[1] = -999999;
	mygraph.edges[j].chain[0] = -999999;
	mygraph.edges[j].probs[0] = -999999;
	e = 1;
      }
      if (!mygraph.edges[j].next) {
	mygraph.edges[j].chain[2] = -999999;
	mygraph.edges[j].probs[2] = -999999;
	mygraph.edges[j].chain[0] = -999999;
	mygraph.edges[j].probs[0] = -999999;
	e = 2;
      }

      if (!e) 
	fprintf(stderr, "WARNING, edge %d is not set!\n", j);
    }
    //    fprintf(stderr, "unclean!\n");
    //      fprintf(stderr,"a: %d %f %f %f\n", mygraph.edges[j].probs[0], mygraph.edges[j].probs[1], mygraph.edges[j].probs[2]);
  }
  //  exit(2);
}

int main(int argc, char** argv) {
  FILE* alignfile;
  int sorted_size;
  int i;
  int first_param = 1;
  char* score_file = NULL;
  
  if (argc < 5) {
    fprintf(stderr,
      "Usage: %s [-score score_file] specie1 specie2 align_file outlier(s)\n",
      argv[0]);
    return 1;
  }
  if (strcasecmp(argv[1], "-score") == 0) {
    score_file = argv[2];
    first_param = 3;
  } else {
    score_file = (char*)ck_malloc(sizeof(SCOREFILE));
    strcpy(score_file, SCOREFILE);
  }
  read_input (argc, argv, first_param, score_file);
  if (!(alignfile = fopen(argv[3 + first_param - 1], "r"))) {
    fprintf(stderr, "Couldn't open aligns file %s\n",
      argv[3 + first_param - 1]);
  }
  initialize_graph(alignfile);
  sorted_aligns = (int*) ck_malloc(sizeof(int) * (mygraph.numaligns+1));
  init_filers();
  for (i = 0; i < 2; i++) {
    sorted_size = sort_graph(i+1); //sort aligns on genome
    fprintf(stderr, "Sorted size1 = %d\n", sorted_size);
    setup_edges(i, sorted_size); // add active edges to all connections
    fprintf(stderr, "donesetup = %d\n", sorted_size);    
#ifdef CLIENT
    while (get_weights(i, sorted_size)) // give them weights
      ;
#endif
  }
#ifdef CLIENT
  dump_weights();
#else
  read_weights();
  solve_flow();  //do the flow 
  output_tigs(); //output pieces of the ancestor
#endif
  return 0;
}
